@@ -1,4 +1,6 @@
 import asyncio
+import json
+import logging
 import os
 from typing import Any, Dict, List, Literal, Optional, Union
 
@@ -11,6 +13,8 @@ from promptlayer.templates import AsyncTemplateManager, TemplateManager
 from promptlayer.track import AsyncTrackManager, TrackManager
 from promptlayer.types.prompt_template import PromptTemplate
 from promptlayer.utils import (
+    RERAISE_ORIGINAL_EXCEPTION,
+    _get_workflow_workflow_id_or_name,
     arun_workflow_request,
     astream_response,
     atrack_request,
@@ -20,6 +24,29 @@ from promptlayer.utils import (
     util_log_request,
 )
 
+logger = logging.getLogger(__name__)
+
+
+def is_workflow_results_dict(obj: Any) -> bool:
+    if not isinstance(obj, dict):
+        return False
+
+    required_keys = {
+        "status",
+        "value",
+        "error_message",
+        "raw_error_message",
+        "is_output_node",
+    }
+
+    for val in obj.values():
+        if not isinstance(val, dict):
+            return False
+        if not required_keys.issubset(val.keys()):
+            return False
+
+    return True
+
 
 class PromptLayer(PromptLayerMixin):
     def __init__(
@@ -39,9 +66,7 @@ class PromptLayer(PromptLayerMixin):
         self.api_key = api_key
         self.templates = TemplateManager(api_key)
         self.group = GroupManager(api_key)
-        self.tracer_provider, self.tracer = self._initialize_tracer(
-            api_key, enable_tracing
-        )
+        self.tracer_provider, self.tracer = self._initialize_tracer(api_key, enable_tracing)
         self.track = TrackManager(api_key)
 
     def __getattr__(
@@ -51,24 +76,17 @@ class PromptLayer(PromptLayerMixin):
         if name == "openai":
             import openai as openai_module
 
-            openai = PromptLayerBase(
-                openai_module,
-                function_name="openai",
-                api_key=self.api_key,
-                tracer=self.tracer,
-            )
-            return openai
+            return PromptLayerBase(openai_module, function_name="openai", api_key=self.api_key, tracer=self.tracer)
         elif name == "anthropic":
             import anthropic as anthropic_module
 
-            anthropic = PromptLayerBase(
+            return PromptLayerBase(
                 anthropic_module,
                 function_name="anthropic",
                 provider_type="anthropic",
                 api_key=self.api_key,
                 tracer=self.tracer,
             )
-            return anthropic
         else:
             raise AttributeError(f"module {__name__} has no attribute {name}")
 
@@ -108,18 +126,22 @@ class PromptLayer(PromptLayerMixin):
         group_id: Union[int, None] = None,
         stream: bool = False,
         pl_run_span_id: Union[str, None] = None,
+        provider: Union[str, None] = None,
+        model: Union[str, None] = None,
     ) -> Dict[str, Any]:
         get_prompt_template_params = self._prepare_get_prompt_template_params(
             prompt_version=prompt_version,
             prompt_release_label=prompt_release_label,
             input_variables=input_variables,
             metadata=metadata,
+            provider=provider,
+            model=model,
         )
         prompt_blueprint = self.templates.get(prompt_name, get_prompt_template_params)
         prompt_blueprint_model = self._validate_and_extract_model_from_prompt_blueprint(
             prompt_blueprint=prompt_blueprint, prompt_name=prompt_name
         )
-        llm_request_params = self._prepare_llm_request_params(
+        llm_data = self._prepare_llm_data(
             prompt_blueprint=prompt_blueprint,
             prompt_template=prompt_blueprint["prompt_template"],
             prompt_blueprint_model=prompt_blueprint_model,
@@ -127,31 +149,36 @@ class PromptLayer(PromptLayerMixin):
             stream=stream,
         )
 
-        response = llm_request_params["request_function"](
-            llm_request_params["prompt_blueprint"], **llm_request_params["kwargs"]
+        # response is just whatever the LLM call returns
+        # streaming=False > Pydantic model instance
+        # streaming=True > generator that yields ChatCompletionChunk pieces as they arrive
+        response = llm_data["request_function"](
+            prompt_blueprint=llm_data["prompt_blueprint"],
+            client_kwargs=llm_data["client_kwargs"],
+            function_kwargs=llm_data["function_kwargs"],
         )
 
         if stream:
             return stream_response(
-                response,
-                self._create_track_request_callable(
-                    request_params=llm_request_params,
+                generator=response,
+                after_stream=self._create_track_request_callable(
+                    request_params=llm_data,
                     tags=tags,
                     input_variables=input_variables,
                     group_id=group_id,
                     pl_run_span_id=pl_run_span_id,
                 ),
-                llm_request_params["stream_function"],
+                map_results=llm_data["stream_function"],
             )
 
         request_log = self._track_request_log(
-            llm_request_params,
+            llm_data,
             tags,
             input_variables,
             group_id,
             pl_run_span_id,
             metadata=metadata,
-            request_response=response.model_dump(),
+            request_response=response.model_dump(mode="json"),
         )
 
         return {
@@ -193,17 +220,21 @@ class PromptLayer(PromptLayerMixin):
         metadata: Union[Dict[str, str], None] = None,
         group_id: Union[int, None] = None,
         stream: bool = False,
+        provider: Union[str, None] = None,
+        model: Union[str, None] = None,
     ) -> Dict[str, Any]:
         _run_internal_kwargs = {
             "prompt_name": prompt_name,
             "prompt_version": prompt_version,
             "prompt_release_label": prompt_release_label,
-            "input_variables": input_variables,
+            "input_variables": input_variables or {},
             "model_parameter_overrides": model_parameter_overrides,
             "tags": tags,
             "metadata": metadata,
             "group_id": group_id,
             "stream": stream,
+            "provider": provider,
+            "model": model,
         }
 
         if self.tracer:
@@ -211,9 +242,7 @@ class PromptLayer(PromptLayerMixin):
                 span.set_attribute("prompt_name", prompt_name)
                 span.set_attribute("function_input", str(_run_internal_kwargs))
                 pl_run_span_id = hex(span.context.span_id)[2:].zfill(16)
-                result = self._run_internal(
-                    **_run_internal_kwargs, pl_run_span_id=pl_run_span_id
-                )
+                result = self._run_internal(**_run_internal_kwargs, pl_run_span_id=pl_run_span_id)
                 span.set_attribute("function_output", str(result))
                 return result
         else:
@@ -221,51 +250,53 @@ class PromptLayer(PromptLayerMixin):
 
     def run_workflow(
         self,
-        workflow_name: str,
+        workflow_id_or_name: Optional[Union[int, str]] = None,
         input_variables: Optional[Dict[str, Any]] = None,
         metadata: Optional[Dict[str, str]] = None,
         workflow_label_name: Optional[str] = None,
-        workflow_version: Optional[
-            int
-        ] = None,  # This is the version number, not the version ID
+        workflow_version: Optional[int] = None,
         return_all_outputs: Optional[bool] = False,
-    ) -> Dict[str, Any]:
+        # `workflow_name` deprecated, kept for backward compatibility only.
+        # Allows `workflow_name` to be passed both as keyword and positional argument
+        # (virtually identical to `workflow_id_or_name`)
+        workflow_name: Optional[str] = None,
+    ) -> Union[Dict[str, Any], Any]:
         try:
             try:
-                # Check if we're inside a running event loop
-                loop = asyncio.get_running_loop()
+                loop = asyncio.get_running_loop()  # Check if we're inside a running event loop
             except RuntimeError:
                 loop = None
 
             if loop and loop.is_running():
                 nest_asyncio.apply()
-                # If there's an active event loop, use `await` directly
-                return asyncio.run(
-                    arun_workflow_request(
-                        workflow_name=workflow_name,
-                        input_variables=input_variables or {},
-                        metadata=metadata,
-                        workflow_label_name=workflow_label_name,
-                        workflow_version_number=workflow_version,
-                        api_key=self.api_key,
-                        return_all_outputs=return_all_outputs,
-                    )
+
+            results = asyncio.run(
+                arun_workflow_request(
+                    workflow_id_or_name=_get_workflow_workflow_id_or_name(workflow_id_or_name, workflow_name),
+                    input_variables=input_variables or {},
+                    metadata=metadata,
+                    workflow_label_name=workflow_label_name,
+                    workflow_version_number=workflow_version,
+                    api_key=self.api_key,
+                    return_all_outputs=return_all_outputs,
                 )
+            )
+
+            if not return_all_outputs and is_workflow_results_dict(results):
+                output_nodes = [node_data for node_data in results.values() if node_data.get("is_output_node")]
+                if not output_nodes:
+                    raise Exception("Output nodes not found: %S", json.dumps(results, indent=4))
+
+                if not any(node.get("status") == "SUCCESS" for node in output_nodes):
+                    raise Exception("None of the output nodes have succeeded", json.dumps(results, indent=4))
+
+            return results
+        except Exception as ex:
+            logger.exception("Error running workflow")
+            if RERAISE_ORIGINAL_EXCEPTION:
+                raise
             else:
-                # If there's no active event loop, use `asyncio.run()`
-                return asyncio.run(
-                    arun_workflow_request(
-                        workflow_name=workflow_name,
-                        input_variables=input_variables or {},
-                        metadata=metadata,
-                        workflow_label_name=workflow_label_name,
-                        workflow_version_number=workflow_version,
-                        api_key=self.api_key,
-                        return_all_outputs=return_all_outputs,
-                    )
-                )
-        except Exception as e:
-            raise Exception(f"Error running workflow: {str(e)}")
+                raise Exception(f"Error running workflow: {str(ex)}") from ex
 
     def log_request(
         self,
@@ -276,6 +307,8 @@ class PromptLayer(PromptLayerMixin):
         output: PromptTemplate,
         request_start_time: float,
         request_end_time: float,
+        # TODO(dmu) MEDIUM: Avoid using mutable defaults
+        # TODO(dmu) MEDIUM: Deprecate and remove this wrapper function?
         parameters: Dict[str, Any] = {},
         tags: List[str] = [],
         metadata: Dict[str, str] = {},
@@ -328,14 +361,10 @@ class AsyncPromptLayer(PromptLayerMixin):
         self.api_key = api_key
         self.templates = AsyncTemplateManager(api_key)
         self.group = AsyncGroupManager(api_key)
-        self.tracer_provider, self.tracer = self._initialize_tracer(
-            api_key, enable_tracing
-        )
+        self.tracer_provider, self.tracer = self._initialize_tracer(api_key, enable_tracing)
         self.track = AsyncTrackManager(api_key)
 
-    def __getattr__(
-        self, name: Union[Literal["openai"], Literal["anthropic"], Literal["prompts"]]
-    ):
+    def __getattr__(self, name: Union[Literal["openai"], Literal["anthropic"], Literal["prompts"]]):
         if name == "openai":
             import openai as openai_module
 
@@ -343,6 +372,7 @@ class AsyncPromptLayer(PromptLayerMixin):
                 openai_module,
                 function_name="openai",
                 api_key=self.api_key,
+                tracer=self.tracer,
             )
             return openai
         elif name == "anthropic":
@@ -353,6 +383,7 @@ class AsyncPromptLayer(PromptLayerMixin):
                 function_name="anthropic",
                 provider_type="anthropic",
                 api_key=self.api_key,
+                tracer=self.tracer,
             )
             return anthropic
         else:
@@ -360,18 +391,20 @@ class AsyncPromptLayer(PromptLayerMixin):
 
     async def run_workflow(
         self,
-        workflow_name: str,
+        workflow_id_or_name: Optional[Union[int, str]] = None,
         input_variables: Optional[Dict[str, Any]] = None,
         metadata: Optional[Dict[str, str]] = None,
         workflow_label_name: Optional[str] = None,
-        workflow_version: Optional[
-            int
-        ] = None,  # This is the version number, not the version ID
+        workflow_version: Optional[int] = None,  # This is the version number, not the version ID
         return_all_outputs: Optional[bool] = False,
-    ) -> Dict[str, Any]:
+        # `workflow_name` deprecated, kept for backward compatibility only.
+        # Allows `workflow_name` to be passed both as keyword and positional argument
+        # (virtually identical to `workflow_id_or_name`)
+        workflow_name: Optional[str] = None,
+    ) -> Union[Dict[str, Any], Any]:
         try:
-            result = await arun_workflow_request(
-                workflow_name=workflow_name,
+            return await arun_workflow_request(
+                workflow_id_or_name=_get_workflow_workflow_id_or_name(workflow_id_or_name, workflow_name),
                 input_variables=input_variables or {},
                 metadata=metadata,
                 workflow_label_name=workflow_label_name,
@@ -379,9 +412,12 @@ class AsyncPromptLayer(PromptLayerMixin):
                 api_key=self.api_key,
                 return_all_outputs=return_all_outputs,
             )
-            return result
-        except Exception as e:
-            raise Exception(f"Error running workflow: {str(e)}")
+        except Exception as ex:
+            logger.exception("Error running workflow")
+            if RERAISE_ORIGINAL_EXCEPTION:
+                raise
+            else:
+                raise Exception(f"Error running workflow: {str(ex)}")
 
     async def run(
         self,
@@ -394,6 +430,8 @@ class AsyncPromptLayer(PromptLayerMixin):
         metadata: Union[Dict[str, str], None] = None,
         group_id: Union[int, None] = None,
         stream: bool = False,
+        provider: Union[str, None] = None,
+        model: Union[str, None] = None,
     ) -> Dict[str, Any]:
         _run_internal_kwargs = {
             "prompt_name": prompt_name,
@@ -405,6 +443,8 @@ class AsyncPromptLayer(PromptLayerMixin):
             "metadata": metadata,
             "group_id": group_id,
             "stream": stream,
+            "provider": provider,
+            "model": model,
         }
 
         if self.tracer:
@@ -412,9 +452,7 @@ class AsyncPromptLayer(PromptLayerMixin):
                 span.set_attribute("prompt_name", prompt_name)
                 span.set_attribute("function_input", str(_run_internal_kwargs))
                 pl_run_span_id = hex(span.context.span_id)[2:].zfill(16)
-                result = await self._run_internal(
-                    **_run_internal_kwargs, pl_run_span_id=pl_run_span_id
-                )
+                result = await self._run_internal(**_run_internal_kwargs, pl_run_span_id=pl_run_span_id)
                 span.set_attribute("function_output", str(result))
                 return result
         else:
@@ -483,7 +521,7 @@ class AsyncPromptLayer(PromptLayerMixin):
             )
             return await atrack_request(**track_request_kwargs)
 
-        return await _track_request
+        return _track_request
 
     async def _track_request_log(
         self,
@@ -520,20 +558,22 @@ class AsyncPromptLayer(PromptLayerMixin):
         group_id: Union[int, None] = None,
         stream: bool = False,
         pl_run_span_id: Union[str, None] = None,
+        provider: Union[str, None] = None,
+        model: Union[str, None] = None,
     ) -> Dict[str, Any]:
         get_prompt_template_params = self._prepare_get_prompt_template_params(
             prompt_version=prompt_version,
             prompt_release_label=prompt_release_label,
             input_variables=input_variables,
             metadata=metadata,
+            provider=provider,
+            model=model,
         )
-        prompt_blueprint = await self.templates.get(
-            prompt_name, get_prompt_template_params
-        )
+        prompt_blueprint = await self.templates.get(prompt_name, get_prompt_template_params)
         prompt_blueprint_model = self._validate_and_extract_model_from_prompt_blueprint(
             prompt_blueprint=prompt_blueprint, prompt_name=prompt_name
         )
-        llm_request_params = self._prepare_llm_request_params(
+        llm_data = self._prepare_llm_data(
             prompt_blueprint=prompt_blueprint,
             prompt_template=prompt_blueprint["prompt_template"],
             prompt_blueprint_model=prompt_blueprint_model,
@@ -542,31 +582,34 @@ class AsyncPromptLayer(PromptLayerMixin):
             is_async=True,
         )
 
-        response = await llm_request_params["request_function"](
-            llm_request_params["prompt_blueprint"], **llm_request_params["kwargs"]
+        response = await llm_data["request_function"](
+            prompt_blueprint=llm_data["prompt_blueprint"],
+            client_kwargs=llm_data["client_kwargs"],
+            function_kwargs=llm_data["function_kwargs"],
         )
 
         if stream:
+            track_request_callable = await self._create_track_request_callable(
+                request_params=llm_data,
+                tags=tags,
+                input_variables=input_variables,
+                group_id=group_id,
+                pl_run_span_id=pl_run_span_id,
+            )
             return astream_response(
                 response,
-                self._create_track_request_callable(
-                    request_params=llm_request_params,
-                    tags=tags,
-                    input_variables=input_variables,
-                    group_id=group_id,
-                    pl_run_span_id=pl_run_span_id,
-                ),
-                llm_request_params["stream_function"],
+                track_request_callable,
+                llm_data["stream_function"],
             )
 
         request_log = await self._track_request_log(
-            llm_request_params,
+            llm_data,
             tags,
             input_variables,
             group_id,
             pl_run_span_id,
             metadata=metadata,
-            request_response=response.model_dump(),
+            request_response=response.model_dump(mode="json"),
         )
 
         return {
